from dataclasses import dataclass
from typing import Sequence

from diamond_miner.defaults import UNIVERSE_SUBSET
from diamond_miner.queries.query import (
    Query,
    links_table,
    prefixes_table,
    probes_table,
    results_table,
)
from diamond_miner.typing import IPNetwork


@dataclass(frozen=True)
class DropTables(Query):
    """Drop the tables associated to a measurement."""

    def statements(
        self, measurement_id: str, subset: IPNetwork = UNIVERSE_SUBSET
    ) -> Sequence[str]:
        return (
            f"DROP TABLE IF EXISTS {results_table(measurement_id)}",
            f"DROP TABLE IF EXISTS {links_table(measurement_id)}",
            f"DROP TABLE IF EXISTS {prefixes_table(measurement_id)}",
            f"DROP TABLE IF EXISTS {probes_table(measurement_id)}",
        )



from dataclasses import dataclass

from diamond_miner.defaults import UNIVERSE_SUBSET
from diamond_miner.queries.fragments import date_time
from diamond_miner.queries.query import Query, StoragePolicy, probes_table
from diamond_miner.typing import IPNetwork


@dataclass(frozen=True)
class CreateProbesTable(Query):
    """Create the table containing the cumulative number of probes sent over the rounds."""

    SORTING_KEY = "probe_protocol, probe_dst_prefix, probe_ttl"

    storage_policy: StoragePolicy = StoragePolicy()

    def statement(
        self, measurement_id: str, subset: IPNetwork = UNIVERSE_SUBSET
    ) -> str:
        return f"""
        CREATE TABLE IF NOT EXISTS {probes_table(measurement_id)}
        (
            probe_protocol         UInt8,
            probe_dst_prefix       IPv6,
            probe_ttl              UInt8,
            cumulative_probes      UInt32,
            round                  UInt8
        )
        ENGINE MergeTree
        ORDER BY ({self.SORTING_KEY})
        TTL {date_time(self.storage_policy.archive_on)} TO VOLUME '{self.storage_policy.archive_to}'
        SETTINGS storage_policy = '{self.storage_policy.name}'
        """



from dataclasses import dataclass, fields
from typing import Sequence

from diamond_miner.defaults import (
    DEFAULT_PREFIX_LEN_V4,
    DEFAULT_PREFIX_LEN_V6,
    UNIVERSE_SUBSET,
)
from diamond_miner.queries import (
    CreateLinksTable,
    CreatePrefixesTable,
    CreateProbesTable,
    CreateResultsTable,
)
from diamond_miner.queries.query import Query, StoragePolicy
from diamond_miner.typing import IPNetwork


@dataclass(frozen=True)
class CreateTables(Query):
    """Create the tables necessary for a measurement."""

    prefix_len_v4: int = DEFAULT_PREFIX_LEN_V4
    prefix_len_v6: int = DEFAULT_PREFIX_LEN_V6
    storage_policy: StoragePolicy = StoragePolicy()

    def statements(
        self, measurement_id: str, subset: IPNetwork = UNIVERSE_SUBSET
    ) -> Sequence[str]:
        all_params = {field.name: getattr(self, field.name) for field in fields(self)}
        # Only CreateResultsTable accepts these parameters.
        params = {
            x: y
            for x, y in all_params.items()
            if x not in ["prefix_len_v4", "prefix_len_v6"]
        }
        return (
            *CreateResultsTable(**all_params).statements(measurement_id, subset),
            *CreateLinksTable(**params).statements(measurement_id, subset),
            *CreatePrefixesTable(**params).statements(measurement_id, subset),
            *CreateProbesTable(**params).statements(measurement_id, subset),
        )



import math
from hashlib import sha256
from typing import Iterator, List, Union
from uuid import UUID

from diamond_miner import __version__
from diamond_miner.defaults import PROTOCOLS


def format_addr(addr: str) -> str:
    if addr.startswith("::ffff:"):
        return addr[7:]
    return addr


def measurement_id_from_name(name: str) -> int:
    m = sha256()
    m.update(name.encode("utf-8"))
    return int.from_bytes(m.digest()[:4], "little")


def probe_id_from_uuid(uuid: Union[UUID, str]) -> int:
    m = sha256()
    m.update(str(uuid).encode())
    return int.from_bytes(m.digest()[:4], "little")


def from_ripe_atlas(traceroute: dict) -> dict:
    """
    >>> from diamond_miner.test import url
    >>> from diamond_miner.queries import GetTraceroutes
    >>> rows = GetTraceroutes().execute(url, 'test_nsdi_example')
    >>> row = sorted(rows, key=lambda x: x["probe_dst_addr"])[0]
    >>> atlas = to_ripe_atlas("measurement", "agent", **row, ipv4_mapped=True)
    >>> res = from_ripe_atlas(atlas)
    >>> res == row
    True
    """
    probe_protocol = PROTOCOLS[traceroute["proto"].lower()]
    probe_src_addr = traceroute["src_addr"]
    probe_dst_addr = traceroute["dst_addr"]
    probe_src_port = traceroute["paris_id"]
    replies = []

    for result in traceroute["result"]:
        for result_ in result["result"]:
            reply_mpls_labels = []
            for obj in result_.get("icmpext", {}).get("obj", []):
                for entry in obj.get("mpls", []):
                    reply_mpls_labels.append(
                        (entry["label"], entry["exp"], entry["s"], entry["ttl"])
                    )

            replies.append(
                [
                    0,
                    result["hop"],
                    result_["ttl"],
                    result_["size"],
                    reply_mpls_labels,
                    result_["from"],
                    result_["rtt"],
                ]
            )

    return {
        "probe_protocol": probe_protocol,
        "probe_src_addr": probe_src_addr,
        "probe_dst_addr": probe_dst_addr,
        "probe_src_port": probe_src_port,
        "replies": replies,
    }


def to_ripe_atlas(
    measurement_uuid: Union[UUID, str],
    agent_uuid: Union[UUID, str],
    probe_protocol: int,
    probe_src_addr: str,
    probe_dst_addr: str,
    probe_src_port: int,
    replies: List[dict],
    *,
    ipv4_mapped: bool = False,
) -> dict:
    """
    >>> from diamond_miner.test import url
    >>> from diamond_miner.queries import GetTraceroutes
    >>> rows = GetTraceroutes().execute(url, 'test_nsdi_example')
    >>> rows = sorted(rows, key=lambda x: x["probe_dst_addr"])
    >>> atlas = to_ripe_atlas("measurement", "agent", **rows[0])
    >>> atlas["src_addr"]
    '100.0.0.1'
    >>> atlas["dst_addr"]
    '200.0.0.0'
    >>> atlas["result"][0]
    {'hop': 1, 'result': [{'from': '150.0.1.1', 'rtt': 0, 'size': 0, 'ttl': 250}]}
    >>> atlas["result"][3]
    {'hop': 4, 'result': [{'from': '150.0.6.1', 'rtt': 0, 'size': 0, 'ttl': 250}]}
    """
    af = 4 if probe_dst_addr.startswith("::ffff:") else 6
    if ipv4_mapped:
        probe_src_addr_str = probe_src_addr
        probe_dst_addr_str = probe_dst_addr
    else:
        probe_src_addr_str = format_addr(probe_src_addr)
        probe_dst_addr_str = format_addr(probe_dst_addr)
    start_timestamp = math.inf
    end_timestamp = -math.inf
    results = []

    for (
        capture_timestamp,
        probe_ttl,
        reply_ttl,
        reply_size,
        reply_mpls_labels,
        reply_src_addr,
        rtt,
    ) in replies:
        start_timestamp = min(start_timestamp, capture_timestamp)
        end_timestamp = max(end_timestamp, capture_timestamp)
        result = {
            "from": reply_src_addr if ipv4_mapped else format_addr(reply_src_addr),
            "rtt": rtt,
            "size": reply_size,
            "ttl": reply_ttl,
        }
        if reply_mpls_labels:
            result["icmpext"] = {
                "obj": [
                    {
                        "mpls": [
                            {"exp": exp, "label": label, "s": s, "ttl": ttl}
                            for (label, exp, s, ttl) in reply_mpls_labels
                        ]
                    }
                ]
            }
        results.append({"hop": probe_ttl, "result": [result]})

    return {
        "af": af,
        "fw": 1,  # TODO: Iris version (must be int)?
        "msm_name": f"{measurement_uuid}__{agent_uuid}",
        "msm_id": measurement_id_from_name(f"{measurement_uuid}__{agent_uuid}"),
        "prb_id": probe_id_from_uuid(agent_uuid),
        "dst_addr": probe_dst_addr_str,
        "dst_name": probe_dst_addr_str,
        "src_addr": probe_src_addr_str,
        "from": probe_src_addr_str,
        "paris_id": probe_src_port,
        "proto": str(PROTOCOLS[probe_protocol]).upper(),
        "result": results,
        "size": 0,  # TODO: Can we know this? Infer from protocol?
        "timestamp": start_timestamp,
        "endtime": end_timestamp,
        "stored_timestamp": end_timestamp,
        "mver": __version__,
        "type": "traceroute",
    }


def unfold(
    probe_protocol: int,
    probe_src_addr: str,
    probe_dst_addr: str,
    probe_src_port: int,
    replies: List[dict],
) -> Iterator[tuple]:
    # TODO: Return dict instead?
    """
    >>> from diamond_miner.test import url
    >>> from diamond_miner.queries import GetTraceroutes
    >>> rows = GetTraceroutes().execute(url, 'test_nsdi_example')
    >>> rows = sorted(rows, key=lambda x: x["probe_dst_addr"])
    >>> next(unfold(**rows[0]))
    (0, 1, '::ffff:100.0.0.1', '::ffff:200.0.0.0', 24000, 0, 1, 0, '::ffff:150.0.1.1', 1, 11, 0, 250, 0, [], 0, 1)
    """
    for (
        capture_timestamp,
        probe_ttl,
        reply_ttl,
        reply_size,
        reply_mpls_labels,
        reply_src_addr,
        rtt,
    ) in replies:
        yield (
            capture_timestamp,
            probe_protocol,
            probe_src_addr,
            probe_dst_addr,
            probe_src_port,
            0,
            probe_ttl,
            0,
            reply_src_addr,
            1,  # TODO: Handle ICMPv6
            11,  # TODO: Handle ICMPv6 Echo Reply
            0,  # TODO: Handle ICMPv6 Echo Reply
            reply_ttl,
            reply_size,
            reply_mpls_labels,
            rtt,
            1,
        )



from dataclasses import dataclass
from typing import List

from diamond_miner.defaults import UNIVERSE_SUBSET
from diamond_miner.queries import GetInvalidPrefixes
from diamond_miner.queries.query import ResultsQuery, results_table
from diamond_miner.typing import IPNetwork
from diamond_miner.utilities import common_parameters


@dataclass(frozen=True)
class GetNodes(ResultsQuery):
    """
    Return all the discovered nodes.

    >>> from diamond_miner.test import url
    >>> nodes = GetNodes(include_probe_ttl=True).execute(url, 'test_nsdi_example')
    >>> sorted((node["probe_ttl"], node["reply_src_addr"]) for node in nodes)
    [(1, '::ffff:150.0.1.1'), (2, '::ffff:150.0.2.1'), (2, '::ffff:150.0.3.1'), (3, '::ffff:150.0.4.1'), (3, '::ffff:150.0.5.1'), (3, '::ffff:150.0.7.1'), (4, '::ffff:150.0.6.1')]
    >>> nodes = GetNodes(filter_invalid_prefixes=False).execute(url, 'test_invalid_prefixes')
    >>> sorted(node["reply_src_addr"] for node in nodes)
    ['::ffff:150.0.0.1', '::ffff:150.0.0.2', '::ffff:150.0.1.1', '::ffff:150.0.1.2', '::ffff:150.0.2.1', '::ffff:150.0.2.2', '::ffff:150.0.2.3']
    >>> nodes = GetNodes(filter_invalid_prefixes=True).execute(url, 'test_invalid_prefixes')
    >>> sorted(node["reply_src_addr"] for node in nodes)
    ['::ffff:150.0.0.1', '::ffff:150.0.0.2', '::ffff:150.0.2.3']
    """

    filter_invalid_prefixes: bool = False
    "If true, exclude nodes from prefixes with amplification or loops."

    include_probe_ttl: bool = False
    "If true, include the TTL at which `reply_src_addr` was seen."

    def columns(self) -> List[str]:
        columns = ["reply_src_addr"]
        if self.include_probe_ttl:
            columns.insert(0, "probe_ttl")
        return columns

    def statement(
        self, measurement_id: str, subset: IPNetwork = UNIVERSE_SUBSET
    ) -> str:
        if self.filter_invalid_prefixes:
            invalid_prefixes_query = GetInvalidPrefixes(
                **common_parameters(self, GetInvalidPrefixes)
            )
            prefix_filter = f"""
                    probe_dst_prefix NOT IN ({invalid_prefixes_query.statement(measurement_id, subset)})
                    """
        else:
            prefix_filter = "1"
        return f"""
        SELECT DISTINCT {','.join(self.columns())}
        FROM {results_table(measurement_id)}
        WHERE {self.filters(subset)} AND {prefix_filter}
        """



from dataclasses import dataclass

from diamond_miner.defaults import UNIVERSE_SUBSET
from diamond_miner.queries.fragments import date_time
from diamond_miner.queries.query import Query, StoragePolicy, links_table
from diamond_miner.typing import IPNetwork


@dataclass(frozen=True)
class CreateLinksTable(Query):
    """Create the links table containing one line per (flow, link) pair."""

    storage_policy: StoragePolicy = StoragePolicy()

    def statement(
        self, measurement_id: str, subset: IPNetwork = UNIVERSE_SUBSET
    ) -> str:
        assert subset == UNIVERSE_SUBSET, "subset not allowed for this query"
        return f"""
        CREATE TABLE IF NOT EXISTS {links_table(measurement_id)}
        (
            probe_protocol    UInt8,
            probe_src_addr    IPv6,
            probe_dst_prefix  IPv6,
            probe_dst_addr    IPv6,
            probe_src_port    UInt16,
            probe_dst_port    UInt16,
            near_round        UInt8,
            far_round         UInt8,
            near_ttl          UInt8,
            far_ttl           UInt8,
            near_addr         IPv6,
            far_addr          IPv6,
            -- Materialized columns
            is_destination    UInt8 MATERIALIZED (near_addr = probe_dst_addr) OR (far_addr = probe_dst_addr),
            is_inter_round    UInt8 MATERIALIZED near_round != far_round,
            is_partial        UInt8 MATERIALIZED near_addr = toIPv6('::') OR far_addr = toIPv6('::'),
            is_virtual        UInt8 MATERIALIZED near_addr = toIPv6('::') AND far_addr = toIPv6('::')
        )
        ENGINE MergeTree
        ORDER BY (
            probe_protocol,
            probe_src_addr,
            probe_dst_prefix,
            probe_dst_addr,
            probe_src_port,
            probe_dst_port
        )
        TTL {date_time(self.storage_policy.archive_on)} TO VOLUME '{self.storage_policy.archive_to}'
        SETTINGS storage_policy = '{self.storage_policy.name}'
        """



import time
from dataclasses import fields
from logging import Logger
from types import TracebackType
from typing import Any, Dict, Optional, Type


def common_parameters(from_dataclass: Any, to_dataclass: Any) -> Dict[str, Any]:
    to_params = {field.name for field in fields(to_dataclass)}
    return {
        field.name: getattr(from_dataclass, field.name)
        for field in fields(from_dataclass)
        if field.name in to_params
    }


class Timer:
    """A very simple timer for profiling code blocks."""

    start_time = None
    total_time = 0

    def start(self) -> None:
        self.start_time = time.time_ns()

    def stop(self) -> None:
        if self.start_time:
            self.total_time += time.time_ns() - self.start_time
            self.start_time = None

    def clear(self) -> None:
        self.start_time = None
        self.total_time = 0

    @property
    def total_ms(self) -> float:
        return self.total_time / 10 ** 6


class LoggingTimer:
    """A very simple timer for logging the execution time of code blocks."""

    def __init__(self, logger: Logger, prefix: str = ""):
        self.logger = logger
        self.prefix = prefix
        self.timer = Timer()

    def __enter__(self) -> None:
        self.logger.info(self.prefix)
        self.timer.clear()
        self.timer.start()

    def __exit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        self.timer.stop()
        self.logger.info("%s time_ms=%s", self.prefix, self.timer.total_ms)



import os
from concurrent.futures import ThreadPoolExecutor, as_completed
from dataclasses import dataclass
from datetime import datetime
from functools import reduce
from typing import Any, Dict, Iterable, Iterator, List, Optional, Sequence, Tuple, Union

import orjson
import requests

from diamond_miner.defaults import UNIVERSE_SUBSET
from diamond_miner.logger import logger
from diamond_miner.queries.fragments import (
    and_,
    eq,
    geq,
    ip_eq,
    ip_in,
    leq,
    lt,
    not_,
    or_,
)
from diamond_miner.typing import IPNetwork
from diamond_miner.utilities import LoggingTimer


def links_table(measurement_id: str) -> str:
    return f"links__{measurement_id}".replace("-", "_")


def prefixes_table(measurement_id: str) -> str:
    return f"prefixes__{measurement_id}".replace("-", "_")


def probes_table(measurement_id: str) -> str:
    return f"probes__{measurement_id}".replace("-", "_")


def results_table(measurement_id: str) -> str:
    return f"results__{measurement_id}".replace("-", "_")


@dataclass(frozen=True)
class StoragePolicy:
    """
    See
    - https://clickhouse.com/docs/en/engines/table-engines/mergetree-family/mergetree/#table_engine-mergetree-ttl
    - https://clickhouse.com/docs/en/engines/table-engines/mergetree-family/mergetree/#table_engine-mergetree-multiple-volumes
    """

    name: str = "default"
    """Name of the ClickHouse storage policy to use for the table."""
    archive_to: str = "default"
    """Name of the ClickHouse archive volume."""
    archive_on: datetime = datetime(2100, 1, 1)
    """Date at which the table will be moved to the archive volume."""


@dataclass(frozen=True)
class Query:
    @property
    def name(self) -> str:
        return self.__class__.__name__

    def statement(
        self, measurement_id: str, subset: IPNetwork = UNIVERSE_SUBSET
    ) -> str:
        # As a query user, prefer calling ``statements`` instead of ``statement`` as there
        # is no guarantees that the query will implement this method and return a single statement.
        raise NotImplementedError

    def statements(
        self, measurement_id: str, subset: IPNetwork = UNIVERSE_SUBSET
    ) -> Sequence[str]:
        # Override this method if you want your query to return multiple statements.
        return (self.statement(measurement_id, subset),)

    def execute(self, *args: Any, **kwargs: Any) -> List[dict]:
        return list(self.execute_iter(*args, **kwargs))

    def execute_iter(
        self,
        url: str,
        measurement_id: str,
        *,
        data: Optional[Any] = None,
        limit: Optional[Tuple[int, int]] = None,
        subsets: Iterable[IPNetwork] = (UNIVERSE_SUBSET,),
        timeout: Optional[Tuple[int, int]] = (5, 3600),
    ) -> Iterator[dict]:
        for subset in subsets:
            for i, statement in enumerate(self.statements(measurement_id, subset)):
                with LoggingTimer(
                    logger,
                    f"query={self.name}#{i} measurement_id={measurement_id} subset={subset} limit={limit}",
                ):
                    params: Dict[str, Union[int, str]] = {
                        "default_format": "JSONEachRow",
                        "limit": limit[0] if limit else 0,
                        "offset": limit[1] if limit else 0,
                        "output_format_json_quote_64bit_integers": 0,
                        "query": statement,
                    }
                    r = requests.post(
                        url,
                        headers={"Accept-encoding": "gzip"},
                        params=params,
                        data=data,
                        stream=True,
                        timeout=timeout,
                    )
                    try:
                        for line in r.iter_lines(chunk_size=2 ** 20):
                            yield orjson.loads(line)
                    except orjson.JSONDecodeError as e:
                        raise RuntimeError(f"Invalid database response: {line}") from e

    def execute_concurrent(
        self,
        url: str,
        measurement_id: str,
        *,
        subsets: Iterable[IPNetwork] = (UNIVERSE_SUBSET,),
        limit: Optional[Tuple[int, int]] = None,
        concurrent_requests: int = (os.cpu_count() or 2) // 2,
    ) -> None:
        logger.info("query=%s concurrent_requests=%s", self.name, concurrent_requests)
        with ThreadPoolExecutor(concurrent_requests) as executor:
            futures = [
                executor.submit(
                    self.execute,
                    url=url,
                    measurement_id=measurement_id,
                    subsets=(subset,),
                    limit=limit,
                )
                for subset in subsets
            ]
            for future in as_completed(futures):
                future.result()


@dataclass(frozen=True)
class LinksQuery(Query):
    filter_inter_round: bool = False
    "If true, exclude links inferred across rounds."

    filter_partial: bool = False
    "If true, exclude partial links: ``('::', node)`` and ``(node, '::')``."

    filter_virtual: bool = False
    "If true, exclude virtual links: ``('::', '::')``."

    near_or_far_addr: Optional[str] = None
    "If specified, keep only the links that contains this IP address."

    probe_protocol: Optional[int] = None
    "If specified, keep only the links inferred from probes sent with this protocol."

    probe_src_addr: Optional[str] = None
    """
    If specified, keep only the links inferred from probes sent by this address.
    This filter is relatively costly (IPv6 comparison on each row).
    """

    round_eq: Optional[int] = None
    "If specified, keep only the links from this round."

    round_leq: Optional[int] = None
    "If specified, keep only the links from this round or before."

    def filters(self, subset: IPNetwork) -> str:
        """``WHERE`` clause common to all queries on the links table."""
        s = []
        if subset != UNIVERSE_SUBSET:
            s += [ip_in("probe_dst_prefix", subset)]
        if self.probe_protocol:
            s += [eq("probe_protocol", self.probe_protocol)]
        if self.probe_src_addr:
            s += [ip_eq("probe_src_addr", self.probe_src_addr)]
        if self.round_eq:
            s += [eq("near_round", self.round_eq), eq("far_round", self.round_eq)]
        if self.round_leq:
            s += [leq("near_round", self.round_leq), leq("far_round", self.round_leq)]
        if self.near_or_far_addr:
            s += [
                or_(
                    ip_eq("near_addr", self.near_or_far_addr),
                    ip_eq("far_addr", self.near_or_far_addr),
                )
            ]
        if self.filter_inter_round:
            s += [not_("is_inter_round")]
        if self.filter_partial:
            s += [not_("is_partial")]
        if self.filter_virtual:
            s += [not_("is_virtual")]
        return reduce(and_, s or ["1"])


@dataclass(frozen=True)
class PrefixesQuery(Query):
    probe_protocol: Optional[int] = None
    "If specified, keep only the links inferred from probes sent with this protocol."

    probe_src_addr: Optional[str] = None
    """
    If specified, keep only the links inferred from probes sent by this address.
    This filter is relatively costly (IPv6 comparison on each row).
    """

    def filters(self, subset: IPNetwork) -> str:
        """``WHERE`` clause common to all queries on the prefixes table."""
        s = []
        if subset != UNIVERSE_SUBSET:
            s += [ip_in("probe_dst_prefix", subset)]
        if self.probe_protocol:
            s += [eq("probe_protocol", self.probe_protocol)]
        if self.probe_src_addr:
            s += [ip_eq("probe_src_addr", self.probe_src_addr)]
        return reduce(and_, s or ["1"])


@dataclass(frozen=True)
class ProbesQuery(Query):
    probe_protocol: Optional[int] = None
    "If specified, keep only probes sent with this protocol."

    probe_ttl_geq: Optional[int] = None
    "If specified, keep only the probes with TTL >= this value."

    probe_ttl_leq: Optional[int] = None
    "If specified, keep only the probes with TTL <= this value."

    round_eq: Optional[int] = None
    "If specified, keep only the probes from this round."

    round_geq: Optional[int] = None
    "If specified, keep only the probes from this round or after."

    round_leq: Optional[int] = None
    "If specified, keep only the probes from this round or before."

    round_lt: Optional[int] = None
    "If specified, keep only the probes from before this round."

    def filters(self, subset: IPNetwork) -> str:
        """``WHERE`` clause common to all queries on the probes table."""
        s = []
        if subset != UNIVERSE_SUBSET:
            s += [ip_in("probe_dst_prefix", subset)]
        if self.probe_protocol:
            s += [eq("probe_protocol", self.probe_protocol)]
        if self.probe_ttl_geq:
            s += [geq("probe_ttl", self.probe_ttl_geq)]
        if self.probe_ttl_leq:
            s += [leq("probe_ttl", self.probe_ttl_leq)]
        if self.round_eq:
            s += [eq("round", self.round_eq)]
        if self.round_geq:
            s += [geq("round", self.round_geq)]
        if self.round_lt:
            s += [lt("round", self.round_lt)]
        if self.round_leq:
            s += [leq("round", self.round_leq)]
        return reduce(and_, s or ["1"])


@dataclass(frozen=True)
class ResultsQuery(Query):
    filter_destination_host: bool = True
    "If true, ignore the replies from the destination host."

    filter_destination_prefix: bool = True
    "If true, ignore the replies from the destination prefix."

    filter_private: bool = True
    "If true, ignore the replies from private IP addresses."

    filter_invalid_probe_protocol: bool = True
    "If true, ignore the replies with probe protocol ≠ ICMP, ICMPv6 or UDP."

    time_exceeded_only: bool = True
    "If true, ignore non ICMP time exceeded replies."

    probe_protocol: Optional[int] = None
    "If specified, keep only the replies to probes sent with this protocol."

    probe_src_addr: Optional[str] = None
    """
    If specified, keep only the replies to probes sent by this address.
    This filter is relatively costly (IPv6 comparison on each row).
    """

    round_eq: Optional[int] = None
    "If specified, keep only the replies from this round."

    round_leq: Optional[int] = None
    "If specified, keep only the replies from this round or before."

    def filters(self, subset: IPNetwork) -> str:
        """``WHERE`` clause common to all queries on the results table."""
        s = []
        if subset != UNIVERSE_SUBSET:
            s += [ip_in("probe_dst_prefix", subset)]
        if self.probe_protocol:
            s += [eq("probe_protocol", self.probe_protocol)]
        if self.probe_src_addr:
            s += [ip_eq("probe_src_addr", self.probe_src_addr)]
        if self.round_eq:
            s += [eq("round", self.round_eq)]
        if self.round_leq:
            s += [leq("round", self.round_leq)]
        if self.filter_destination_host:
            s += [not_("destination_host_reply")]
        if self.filter_destination_prefix:
            s += [not_("destination_prefix_reply")]
        if self.filter_private:
            s += [not_("private_probe_dst_prefix"), not_("private_reply_src_addr")]
        if self.time_exceeded_only:
            s += ["time_exceeded_reply"]
        if self.filter_invalid_probe_protocol:
            s += ["valid_probe_protocol"]
        return reduce(and_, s or ["1"])



from dataclasses import dataclass

from diamond_miner.defaults import UNIVERSE_SUBSET
from diamond_miner.queries.query import ResultsQuery, results_table
from diamond_miner.typing import IPNetwork


@dataclass(frozen=True)
class GetTraceroutes(ResultsQuery):
    """
    Return all the columns from the results table.

    >>> from diamond_miner.test import url
    >>> rows = GetTraceroutes().execute(url, 'test_nsdi_example')
    >>> len(rows)
    27
    """

    def statement(
        self, measurement_id: str, subset: IPNetwork = UNIVERSE_SUBSET
    ) -> str:
        return f"""
        SELECT
            probe_protocol,
            probe_src_addr,
            probe_dst_addr,
            probe_src_port,
            groupArray((
                toUnixTimestamp(capture_timestamp),
                probe_ttl,
                reply_ttl,
                reply_size,
                reply_mpls_labels,
                reply_src_addr,
                rtt
            )) AS replies
        FROM {results_table(measurement_id)}
        WHERE {self.filters(subset)}
        GROUP BY (
            probe_protocol,
            probe_src_addr,
            probe_dst_addr,
            probe_src_port,
            probe_dst_port,
        )
        """



import os
from dataclasses import dataclass
from typing import Iterable, Iterator, Tuple

from diamond_miner.defaults import (
    DEFAULT_FAILURE_RATE,
    DEFAULT_PREFIX_LEN_V4,
    DEFAULT_PREFIX_LEN_V6,
    PROTOCOLS,
    UNIVERSE_SUBSET,
)
from diamond_miner.format import format_ipv6
from diamond_miner.generators.standalone import split_prefix
from diamond_miner.queries import InsertMDAProbes
from diamond_miner.queries.query import Query, probes_table
from diamond_miner.subsets import subsets_for
from diamond_miner.typing import IPNetwork


@dataclass(frozen=True)
class InsertProbes(Query):
    def statement(
        self, measurement_id: str, subset: IPNetwork = UNIVERSE_SUBSET
    ) -> str:
        return f"INSERT INTO {probes_table(measurement_id)} FORMAT JSONCompactEachRow"


def insert_probe_counts(
    url: str,
    measurement_id: str,
    round_: int,
    # prefix (/32 or / 128 if nothing specified), protocol, ttls, n_probes
    prefixes: Iterable[Tuple[str, str, Iterable[int], int]],
    prefix_len_v4: int = DEFAULT_PREFIX_LEN_V4,
    prefix_len_v6: int = DEFAULT_PREFIX_LEN_V6,
) -> None:
    """
    >>> from diamond_miner.test import create_tables, url
    >>> from diamond_miner.queries import GetProbes
    >>> create_tables(url, "test_probe_counts")
    >>> insert_probe_counts(url, "test_probe_counts", 1, [("8.8.0.0/22", "icmp", range(2, 5), 6)])
    >>> rows = sorted(GetProbes(round_eq=1).execute(url, "test_probe_counts"), key=lambda x: x["probe_dst_prefix"])
    >>> len(rows)
    4
    >>> row = rows[0]
    >>> row["probe_dst_prefix"]
    '::ffff:8.8.0.0'
    >>> sorted(row["probes_per_ttl"])
    [[2, 6], [3, 6], [4, 6]]
    >>> row = rows[1]
    >>> row["probe_dst_prefix"]
    '::ffff:8.8.1.0'
    >>> sorted(row["probes_per_ttl"])
    [[2, 6], [3, 6], [4, 6]]
    """

    def gen() -> Iterator[bytes]:
        for prefix, protocol, ttls, n_probes in prefixes:
            protocol = PROTOCOLS[protocol]  # type: ignore
            for af, subprefix, subprefix_size in split_prefix(
                prefix, prefix_len_v4, prefix_len_v6
            ):
                yield "\n".join(
                    f'[{protocol},"{format_ipv6(subprefix)}",{ttl},{n_probes},{round_}]'
                    for ttl in ttls
                ).encode()

    InsertProbes().execute(url, measurement_id, data=gen())


def insert_mda_probe_counts(
    url: str,
    measurement_id: str,
    previous_round: int,
    adaptive_eps: bool = False,
    target_epsilon: float = DEFAULT_FAILURE_RATE,
    concurrent_requests: int = (os.cpu_count() or 2) // 2,
) -> None:
    # TODO: set filter_partial and filter_virtual to false?
    query = InsertMDAProbes(
        adaptive_eps=adaptive_eps,
        round_leq=previous_round,
        filter_partial=True,
        filter_virtual=True,
        filter_inter_round=True,
        target_epsilon=target_epsilon,
    )
    subsets = subsets_for(query, url, measurement_id)
    query.execute_concurrent(
        url, measurement_id, subsets=subsets, concurrent_requests=concurrent_requests
    )



from os import environ

url = environ.get("DIAMOND_MINER_TEST_DATABASE_URL", "http://localhost:8123")


def create_tables(url: str, measurement_id: str) -> None:
    # Avoid circular imports.
    from diamond_miner.queries.create_tables import CreateTables
    from diamond_miner.queries.drop_tables import DropTables

    DropTables().execute(url, measurement_id)
    CreateTables().execute(url, measurement_id)



from dataclasses import asdict, dataclass

from diamond_miner.defaults import UNIVERSE_SUBSET
from diamond_miner.queries import GetPrefixesWithAmplification, GetPrefixesWithLoops
from diamond_miner.queries.query import ResultsQuery, prefixes_table, results_table
from diamond_miner.typing import IPNetwork


@dataclass(frozen=True)
class InsertPrefixes(ResultsQuery):
    """Insert (invalid) prefixes into the prefixes table."""

    def statement(
        self, measurement_id: str, subset: IPNetwork = UNIVERSE_SUBSET
    ) -> str:
        amplification_query = GetPrefixesWithAmplification(**asdict(self)).statement(
            measurement_id, subset
        )
        loops_query = GetPrefixesWithLoops(**asdict(self)).statement(
            measurement_id, subset
        )
        return f"""
        INSERT INTO {prefixes_table(measurement_id)}
        SELECT
            prefixes.probe_protocol,
            prefixes.probe_src_addr,
            prefixes.probe_dst_prefix,
            amplification.has_amplification,
            loops.has_loops
        FROM (
            SELECT DISTINCT probe_protocol, probe_src_addr, probe_dst_prefix
            FROM {results_table(measurement_id)}
            WHERE {self.filters(subset)}
        ) AS prefixes
        FULL OUTER JOIN ({amplification_query}) AS amplification
        ON  prefixes.probe_protocol   = amplification.probe_protocol
        AND prefixes.probe_src_addr   = amplification.probe_src_addr
        AND prefixes.probe_dst_prefix = amplification.probe_dst_prefix
        FULL OUTER JOIN ({loops_query}) AS loops
        ON  prefixes.probe_protocol   = loops.probe_protocol
        AND prefixes.probe_src_addr   = loops.probe_src_addr
        AND prefixes.probe_dst_prefix = loops.probe_dst_prefix
        WHERE prefixes.probe_dst_prefix != toIPv6('::')
        """



from dataclasses import dataclass
from typing import Optional

from diamond_miner.defaults import UNIVERSE_SUBSET
from diamond_miner.queries import CreatePrefixesTable
from diamond_miner.queries.fragments import ip_in
from diamond_miner.queries.query import ResultsQuery, prefixes_table, results_table
from diamond_miner.typing import IPNetwork


@dataclass(frozen=True)
class GetPrefixes(ResultsQuery):
    """
    Return the destination prefixes for which replies have been received.

    >>> from diamond_miner.test import url
    >>> from ipaddress import ip_network
    >>> rows = GetPrefixes().execute(url, 'test_nsdi_example')
    >>> len(rows)
    1
    >>> rows = GetPrefixes().execute(url, 'test_invalid_prefixes')
    >>> len(rows)
    3
    >>> rows = GetPrefixes(reply_src_addr_in=ip_network("150.0.1.0/24")).execute(url, 'test_invalid_prefixes')
    >>> len(rows)
    1
    """

    reply_src_addr_in: Optional[IPNetwork] = None

    def statement(
        self, measurement_id: str, subset: IPNetwork = UNIVERSE_SUBSET
    ) -> str:
        # The prefixes table doesn't contains network information, so we
        # need to join the results table for these filters.
        join_fragment = ""
        if self.reply_src_addr_in:
            join_fragment = f"""
            INNER JOIN (
                SELECT DISTINCT probe_protocol, probe_src_addr, probe_dst_prefix
                FROM {results_table(measurement_id)}
                WHERE {self.filters(subset)} AND {ip_in('reply_src_addr', self.reply_src_addr_in)}
            ) AS results
            ON  prefixes.probe_protocol   = results.probe_protocol
            AND prefixes.probe_src_addr   = results.probe_src_addr
            AND prefixes.probe_dst_prefix = results.probe_dst_prefix
            """
        return f"""
        SELECT probe_dst_prefix, has_amplification, has_loops
        FROM {prefixes_table(measurement_id)} AS prefixes
        {join_fragment}
        ORDER BY {CreatePrefixesTable.SORTING_KEY}
        """



from dataclasses import dataclass
from typing import Optional

from diamond_miner.defaults import UNIVERSE_SUBSET
from diamond_miner.queries.query import Query
from diamond_miner.typing import IPNetwork


@dataclass(frozen=True)
class Count(Query):
    """
    Count the rows of a given query.

    >>> from diamond_miner.test import url
    >>> from diamond_miner.queries.get_nodes import GetNodes
    >>> from diamond_miner.queries.get_links import GetLinks
    >>> Count(query=GetNodes()).execute(url, 'test_nsdi_example')[0]["count()"]
    7
    >>> Count(query=GetLinks()).execute(url, 'test_nsdi_example')[0]["count()"]
    8
    """

    query: Optional[Query] = None
    "The query for which to count the nodes."

    def statement(
        self, measurement_id: str, subset: IPNetwork = UNIVERSE_SUBSET
    ) -> str:
        # `query` must be typed `Optional` since it appears after arguments with default values.
        assert self.query is not None
        return f"SELECT count() FROM ({self.query.statement(measurement_id, subset)})"



"""
*D-Miner is the first Internet-scale system that captures a multipath view of the topology.
By combining and adapting state-of-the-art multipath detection and high speed randomized topology discovery techniques,
D-Miner permits discovery of the Internet’s multipath topology in 2.5 days when probing at 100kpps.* :cite:`DiamondMiner2020`

There are two implementations of Diamond-Miner:

- `diamond-miner-cpp <https://github.com/dioptra-io/diamond-miner-cpp>`_ +
  `diamond-miner-wrapper <https://github.com/dioptra-io/diamond-miner-wrapper>`_,
  the original implementation in C++ that have been used for the NSDI 2020 paper :cite:`DiamondMiner2020`.
  This implementation is not maintained anymore.
- This implementation, `diamond-miner <https://github.com/dioptra-io/diamond-miner>`_, a rewrite of the core algorithm in Python and ClickHouse SQL.
  This implementation is maintained and used in production. It supports IPv4 and IPv6.
"""

__version__ = "0.7.2"
